package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.*;
import dev.pfaff.altargui.function.IndexedFunction;
import dev.pfaff.altargui.handlers.MouseScrollHandler;
import dev.pfaff.altargui.inspector.DebugLayer;
import dev.pfaff.altargui.internal.util.StencilLayoutContext;
import org.jetbrains.annotations.NotNull;

import java.util.List;
import java.util.function.BiConsumer;

import static dev.pfaff.altargui.util.ClipUtil.clipExtent;

public record FiniteScrollListWidget(int count, KeyFunction keyFunction, Builder builder) implements Widget {
	public static final DebugLayer DEBUG_LAYER_PRE_UNSCISSORED = new DebugLayer(FiniteScrollListWidget.class, "pre_unscissored");
	public static final DebugLayer DEBUG_LAYER_PRE_SCISSORED = new DebugLayer(FiniteScrollListWidget.class, "pre_scissored");
	public static final DebugLayer DEBUG_LAYER_POST_UNSCISSORED = new DebugLayer(FiniteScrollListWidget.class, "post_unscissored");

	public static <T> FiniteScrollListWidget create(List<T> items,
													IndexedFunction<T, Key.CallSite> keyFunction,
													IndexedFunction<T, Widget> builder) {
		return new FiniteScrollListWidget(items.size(),
				i -> keyFunction.apply(i, items.get(i)),
				i -> builder.apply(i, items.get(i)));
	}

	@Override
	public @NotNull WidgetInstance instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	@FunctionalInterface
	public interface KeyFunction {
		Key.CallSite apply(int index);
	}

	@FunctionalInterface
	public interface Builder {
		Widget build(int index);
	}

	private static final class State extends WidgetInstance<FiniteScrollListWidget> implements MouseScrollHandler {
		private int count;
		private KeyFunction keyFunction;
		private Builder builder;

		/**
		 * The index of the first item displayed.
		 */
		private int itemI;

		/**
		 * The offset of the first item displayed.
		 */
		private double itemOffset;

		private final WidgetInstanceList children = new WidgetInstanceList(2);

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, FiniteScrollListWidget widget, boolean init) {
			var dirtyStencil = false;
			if (count != widget.count) {
				count = widget.count;
				dirtyStencil = true;
			}
			if (!widget.keyFunction.equals(keyFunction)) {
				keyFunction = widget.keyFunction;
				dirtyStencil = true;
			}
			if (!widget.builder.equals(builder)) {
				builder = widget.builder;
				dirtyStencil = true;
			}

			if (dirtyStencil) markNeedsStencilLayout();
		}

		// TODO: cache built widgets
		@Override
		protected void computeStencilLayout(StateContext context) {
			markNeedsTraceLayout();

			var children = this.children;
			children.clear();

			var builder = this.builder;
			var keyFunction = this.keyFunction;
			int count = this.count;
			final var allocWidth = this.allocatedWidth();
			final var allocHeight = this.allocatedHeight();
			float width = 0;
			var offY = this.itemOffset;
			eachElement: for (int i = this.itemI; i < count && offY < allocHeight; ) {
				var widget = builder.build(i);
				{
					var availHeight = (float)(allocHeight - offY);
					int i_ = i;
					var element = context.getState(this, keyFunction.apply(i), widget);
					var offY_ = offY;
					var layoutFunction = new BiConsumer<WidgetInstance<?>, StateContext>() {
						private double offYSlot = offY_;

						@Override
						public void accept(WidgetInstance<?> e, StateContext ctx) {
							e.setAvailableSize(allocWidth, availHeight);
							e.stencilLayout(ctx);

							float height = e.computedHeight();
							if (i_ == State.this.itemI && i_ == count - 1) {
								if (this.offYSlot < 0.0 && this.offYSlot + height < allocHeight) {
									this.offYSlot = (State.this.itemOffset = Math.max(0, height - allocHeight));
								}
							} else if (this.offYSlot + height <= 0.0) {
								if (i_ + 1 < count) {
									State.this.itemI = i_ + 1;
									State.this.itemOffset += height;
								}
							}
						}
					};
					StencilLayoutContext.updateMountAndLayout(context, this, element, layoutFunction);
					float height = element.computedHeight();
					if (i == this.itemI && offY > 0.0) {
						if (i > 0) {
							this.itemI = --i;
							this.itemOffset -= height;
							offY -= height;
						} else {
							this.itemOffset = 0;
							offY = 0.0;
						}
						continue eachElement;
					}
					offY = layoutFunction.offYSlot;
					children.add(element);
					width = Math.max(width, element.computedWidth());
					offY += height;
				}

				i++;
			}

			setComputedSize(width, allocHeight);
		}

		@Override
		protected void computeTraceLayout() {
			markNeedsScissorLayout();

			var children = this.children;

			final float x = this.allocatedX();
			final float minY = this.allocatedY();

			double y = minY + this.itemOffset;

			for (var child : children) {
				child.setPosition(x, (float)y);
				child.traceLayout();
				y += child.computedHeight();
			}
		}

		@Override
		protected void computeScissorLayout() {
			var clippedWidth = clipExtent(this.clippedX(),
					this.clippedWidth(),
					this.allocatedX(),
					this.computedWidth());
			var clippedHeight = clipExtent(this.clippedY(),
					this.clippedHeight(),
					this.allocatedY(),
					this.computedHeight());
			this.children.forEach(child -> {
				child.setClippedRect(clippedWidth.start(), clippedHeight.start(), clippedWidth.extent(), clippedHeight.extent());
				child.scissorLayout();
			});
		}

		@Override
		protected void computeMount(MountContext context) {
			context.pushScissor(this);
			context.mountList(this.children);
			context.popScissor();
		}

		@Override
		public void uSpecMount(MountContext context) {
			super.uSpecMount(context);
		}

		@Override
		public boolean mouseScrolled(double mouseX, double mouseY, double delta) {
			if (isInside(mouseX, mouseY)) {
				this.itemOffset += MouseScrollHandler.applyStandardDeltaTransform(delta);
				this.markNeedsStencilLayout();
				return true;
			}
			return false;
		}
	}
}
